

var CodeMirror = {};
CodeMirror.defaults = {};

var modes = CodeMirror.modes = {}, mimeModes = CodeMirror.mimeModes = {};

CodeMirror.defineMode = function(name, mode) {
  if (!CodeMirror.defaults.mode && name != "null") CodeMirror.defaults.mode = name;
  if (arguments.length > 2) {
    mode.dependencies = [];
    for (var i = 2; i < arguments.length; ++i) mode.dependencies.push(arguments[i]);
  }
  modes[name] = mode;
};

CodeMirror.defineMIME = function(mime, spec) {
  mimeModes[mime] = spec;
};

CodeMirror.resolveMode = function(spec) {
  if (typeof spec == "string" && mimeModes.hasOwnProperty(spec)) {
    spec = mimeModes[spec];
  } else if (spec && typeof spec.name == "string" && mimeModes.hasOwnProperty(spec.name)) {
    var found = mimeModes[spec.name];
    if (typeof found == "string") found = {name: found};
    spec = createObj(found, spec);
    spec.name = found.name;
  } else if (typeof spec == "string" && /^[\w\-]+\/[\w\-]+\+xml$/.test(spec)) {
    return CodeMirror.resolveMode("application/xml");
  }
  if (typeof spec == "string") return {name: spec};
  else return spec || {name: "null"};
};

CodeMirror.getMode = function(options, spec) {
  var spec = CodeMirror.resolveMode(spec);
  var mfactory = modes[spec.name];
  if (!mfactory) return CodeMirror.getMode(options, "text/plain");
  var modeObj = mfactory(options, spec);
  if (modeExtensions.hasOwnProperty(spec.name)) {
    var exts = modeExtensions[spec.name];
    for (var prop in exts) {
      if (!exts.hasOwnProperty(prop)) continue;
      if (modeObj.hasOwnProperty(prop)) modeObj["_" + prop] = modeObj[prop];
      modeObj[prop] = exts[prop];
    }
  }
  modeObj.name = spec.name;
  if (spec.helperType) modeObj.helperType = spec.helperType;
  if (spec.modeProps) for (var prop in spec.modeProps)
    modeObj[prop] = spec.modeProps[prop];

  return modeObj;
};

CodeMirror.defineMode("null", function() {
  return {token: function(stream) {stream.skipToEnd();}};
});
CodeMirror.defineMIME("text/plain", "null");

var modeExtensions = CodeMirror.modeExtensions = {};
CodeMirror.extendMode = function(mode, properties) {
  var exts = modeExtensions.hasOwnProperty(mode) ? modeExtensions[mode] : (modeExtensions[mode] = {});
  copyObj(properties, exts);
};

CodeMirror.defineExtension = function(name, func) {
  CodeMirror.prototype[name] = func;
};
CodeMirror.defineDocExtension = function(name, func) {
  Doc.prototype[name] = func;
};

var initHooks = [];
CodeMirror.defineInitHook = function(f) {initHooks.push(f);};

var helpers = CodeMirror.helpers = {};
CodeMirror.registerHelper = function(type, name, value) {
  if (!helpers.hasOwnProperty(type)) helpers[type] = CodeMirror[type] = {_global: []};
  helpers[type][name] = value;
};
CodeMirror.registerGlobalHelper = function(type, name, predicate, value) {
  CodeMirror.registerHelper(type, name, value);
  helpers[type]._global.push({pred: predicate, val: value});
};

var copyState = CodeMirror.copyState = function(mode, state) {
  if (state === true) return state;
  if (mode.copyState) return mode.copyState(state);
  var nstate = {};
  for (var n in state) {
    var val = state[n];
    if (val instanceof Array) val = val.concat([]);
    nstate[n] = val;
  }
  return nstate;
};

var startState = CodeMirror.startState = function(mode, a1, a2) {
  return mode.startState ? mode.startState(a1, a2) : true;
};

CodeMirror.innerMode = function(mode, state) {
  while (mode.innerMode) {
    var info = mode.innerMode(state);
    if (!info || info.mode == mode) break;
    state = info.state;
    mode = info.mode;
  }
  return info || {mode: mode, state: state};
};

//*******************

var StringStream = CodeMirror.StringStream = function(string, tabSize) {
  this.pos = this.start = 0;
  this.string = string;
  this.tabSize = tabSize || 8;
  this.lastColumnPos = this.lastColumnValue = 0;
  this.lineStart = 0;
};

StringStream.prototype = {
  eol: function() {return this.pos >= this.string.length;},
  sol: function() {return this.pos == this.lineStart;},
  peek: function() {return this.string.charAt(this.pos) || undefined;},
  next: function() {
    if (this.pos < this.string.length)
      return this.string.charAt(this.pos++);
  },
  eat: function(match) {
    var ch = this.string.charAt(this.pos);
    if (typeof match == "string") var ok = ch == match;
    else var ok = ch && (match.test ? match.test(ch) : match(ch));
    if (ok) {++this.pos; return ch;}
  },
  eatWhile: function(match) {
    var start = this.pos;
    while (this.eat(match)){}
    return this.pos > start;
  },
  eatSpace: function() {
    var start = this.pos;
    while (/[\s\u00a0]/.test(this.string.charAt(this.pos))) ++this.pos;
    return this.pos > start;
  },
  skipToEnd: function() {this.pos = this.string.length;},
  skipTo: function(ch) {
    var found = this.string.indexOf(ch, this.pos);
    if (found > -1) {this.pos = found; return true;}
  },
  backUp: function(n) {this.pos -= n;},
  column: function() {
    if (this.lastColumnPos < this.start) {
      this.lastColumnValue = countColumn(this.string, this.start, this.tabSize, this.lastColumnPos, this.lastColumnValue);
      this.lastColumnPos = this.start;
    }
    return this.lastColumnValue - (this.lineStart ? countColumn(this.string, this.lineStart, this.tabSize) : 0);
  },
  indentation: function() {
    return countColumn(this.string, null, this.tabSize) -
      (this.lineStart ? countColumn(this.string, this.lineStart, this.tabSize) : 0);
  },
  match: function(pattern, consume, caseInsensitive) {
    if (typeof pattern == "string") {
      var cased = function(str) {return caseInsensitive ? str.toLowerCase() : str;};
      var substr = this.string.substr(this.pos, pattern.length);
      if (cased(substr) == cased(pattern)) {
        if (consume !== false) this.pos += pattern.length;
        return true;
      }
    } else {
      var match = this.string.slice(this.pos).match(pattern);
      if (match && match.index > 0) return null;
      if (match && consume !== false) this.pos += match[0].length;
      return match;
    }
  },
  current: function(){return this.string.slice(this.start, this.pos);},
  hideFirstChars: function(n, inner) {
    this.lineStart += n;
    try { return inner(); }
    finally { this.lineStart -= n; }
  }
};

var countColumn = CodeMirror.countColumn = function(string, end, tabSize, startIndex, startValue) {
  if (end == null) {
    end = string.search(/[^\s\u00a0]/);
    if (end == -1) end = string.length;
  }
  for (var i = startIndex || 0, n = startValue || 0;;) {
    var nextTab = string.indexOf("\t", i);
    if (nextTab < 0 || nextTab >= end)
      return n + (end - i);
    n += nextTab - i;
    n += tabSize - (n % tabSize);
    i = nextTab + 1;
  }
};

function parseLine(pText, pMode, pHandler, pState) {
  if(!pState && pMode.startState) pState = pMode.startState();
  var flattenSpans = pMode.flattenSpans;
  if (flattenSpans == null) flattenSpans = true;
  var curText = "", curStyle = null;
  var stream = new StringStream(pText), style;
  if (pText == "" && pMode.blankLine) pMode.blankLine(pState);
  while (!stream.eol()) {
    style = pMode.token(stream, pState);
    var substr = stream.current();
    stream.start = stream.pos;
    if (!flattenSpans || curStyle != style) {
      if (curText) pHandler.token(curText, curStyle);
      curText = substr; curStyle = style;
    } else curText = curText + substr;
  }
  if (curText) pHandler.token(curText, curStyle);
}


function parseText(pText, pMimeType, pJsonOptions, pHandler, pMaxLines) {
	var vLines = pText.split(/\r\n?|\n/);
	var vMode = CodeMirror.getMode(pJsonOptions ? JSON.parse(pJsonOptions):{}, pMimeType);
	var vState = vMode.startState ? vMode.startState() : null;
	for(var i=0; i < vLines.length && pMaxLines != 0; i++) {
		pHandler.startLine();
		parseLine(vLines[i], vMode, pHandler, vState);
		pHandler.endLine();
		if (pMaxLines>0) pMaxLines--;
	}
}